"""Module for file naming utilities"""
# Standard
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from importlib import metadata
import os
import re
from types import SimpleNamespace
from typing import Union
from pathlib import Path
# Installed
from cloudpathlib import AnyPath, CloudPath, S3Path
# Local
from libera_utils.time import PRINTABLE_TS_FORMAT, NUMERIC_DOY_TS_FORMAT

REVISION_TS_FORMAT = f"R{NUMERIC_DOY_TS_FORMAT}"  # Just adds an r in front

SPK_REGEX = re.compile(r"^LIBERA_(?P<spk_object>JPSS)"
                       r"_(?P<version>V[0-9]*-[0-9]*-[0-9]*(RC[0-9])?)"
                       r"_(?P<utc_start>[0-9]{8}T[0-9]{6})"
                       r"_(?P<utc_end>[0-9]{8}T[0-9]{6})"
                       r"_(?P<revision>R[0-9]{11})"
                       r"\.bsp$")

CK_REGEX = re.compile(r"^LIBERA_(?P<ck_object>JPSS|AZROT|ELSCAN)"
                      r"_(?P<version>V[0-9]*-[0-9]*-[0-9]*(RC[0-9])?)"
                      r"_(?P<utc_start>[0-9]{8}T[0-9]{6})"
                      r"_(?P<utc_end>[0-9]{8}T[0-9]{6})"
                      r"_(?P<revision>R[0-9]{11})"
                      r"\.bc$")

# L0 filename format determined by EDOS Production Data Set and Construction Record filenaming conventions
LIBERA_L0_REGEX = re.compile(r"^(?P<id_char>P|X)"
                             r"(?P<scid>[0-9]{3})"
                             r"(?P<first_apid>[0-9]{4})"
                             # In some cases at least, the last character of the fill field specifies a time (T)
                             # or session (S) based product. e.g. VIIRSSCIENCEAT
                             r"(?P<fill>.{14})"
                             r"(?P<created_time>[0-9]{11})"
                             r"(?P<numeric_id>[0-9])"
                             r"(?P<file_number>[0-9]{2})"
                             r".(?P<extension>PDR|PDS)"
                             r"(?P<signal>.XFR)?$")

LIBERA_DATA_PRODUCT_REGEX = re.compile(r"^LIBERA_(?P<data_level>L1B|L2)"
                                  r"_(?P<product_name>[^_]*)"
                                  r"_(?P<version>V[0-9]*-[0-9]*-[0-9]*(RC[0-9])?)"
                                  r"_(?P<utc_start>[0-9]{8}T[0-9]{6})"
                                  r"_(?P<utc_end>[0-9]{8}T[0-9]{6})"
                                  r"_(?P<revision>R[0-9]{11})"
                                  r"\.(?P<extension>nc|h5)$")

MANIFEST_FILE_REGEX = re.compile(r"^LIBERA"
                                 r"_(?P<manifest_type>INPUT|OUTPUT)"
                                 r"_MANIFEST"
                                 r"_(?P<created_time>[0-9]{8}T[0-9]{6})"
                                 r"\.json")


class DataLevel(Enum):
    """Data product level"""
    L0 = "L0"
    L1B = 'L1B'
    L2 = 'L2'


class ManifestType(Enum):
    """Enumerated legal manifest type values"""
    INPUT = 'INPUT'
    input = INPUT
    OUTPUT = 'OUTPUT'
    output = OUTPUT


class AnyFilename:
    """Polymorphic class for creating a Filename object"""

    def __new__(cls, *args, **kwargs) -> 'AbstractValidFilename':
        for CandidateClass in (L0Filename, AttitudeKernelFilename,
                               EphemerisKernelFilename, LiberaDataProductFilename,
                               ManifestFilename):
            try:
                filename = CandidateClass(*args, **kwargs)
                return filename
            except ValueError:
                continue

        raise ValueError(f"Unable to create a valid filename from {args}. "
                         "Are you sure this is a valid Libera file name?")


class AbstractValidFilename(ABC):
    """Composition of a CloudPath/Path instance with some methods to perform
    regex validation on filenames
    """
    _regex: re.Pattern
    _fmt: str
    _required_parts: tuple

    def __init__(self, *args, **kwargs):
        self.path = AnyPath(*args, **kwargs)

    def __str__(self):
        return str(self.path)

    def __eq__(self, other):
        if self.path == other.path and self.filename_parts == other.filename_parts:
            return True
        return False

    @property
    def path(self):
        """Property containing the file path"""
        return self._path

    @path.setter
    def path(self, new_path: str or Path or CloudPath):
        if isinstance(new_path, str):
            new_path = AnyPath(new_path)
        self.regex_match(new_path)  # validates against regex pattern
        self._path: CloudPath or Path = AnyPath(new_path)

    @property
    def filename_parts(self):
        """Property that contains a namespace of filename parts"""
        return self._parse_filename_parts()

    @property
    @abstractmethod
    def archive_prefix(self):
        """Property that contains the generated prefix used for archiving, when applicable"""
        raise NotImplementedError()

    @classmethod
    def _check_required_parts(cls, local_vars: dict):
        """Checks for the presence of required filename parts

        Parameters
        ----------
        local_vars : dict
            Dictionary of variables passed, created by a call to locals()
        """
        print(local_vars)
        missing = [req for req in cls._required_parts if local_vars[req] is None]
        if missing:
            raise ValueError(
                f"Missing required keyword argument(s) to {cls.__name__}.from_filename_parts: {', '.join(missing)}.")

    @classmethod
    @abstractmethod
    def from_filename_parts(cls,
                            *args,
                            basepath: str or Path = None):
        """Abstract method that must be implemented to provide hinting for required parts"""
        raise NotImplementedError()

    @classmethod
    def _from_filename_parts(cls,
                             basepath: str or Path = None,
                             **parts):
        """Create instance from filename parts.

        The part arg names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath or prefix.
        parts :
            Passed directly to _format_filename_parts

        Returns
        -------
        : cls
        """
        filename = cls._format_filename_parts(**parts)
        if basepath is not None:
            return cls(os.path.join(basepath, filename))
        return cls(filename)

    @classmethod
    @abstractmethod
    def _format_filename_parts(cls, **parts):
        """Format parts into a filename

        Note: When this is implemented by concrete classes, **parts becomes a set of explicitly named arguments
        """
        raise NotImplementedError()

    @abstractmethod
    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        # Do stuff to parse the elements of d into a SimpleNamespace
        raise NotImplementedError()

    @staticmethod
    def _calculate_applicable_time(start: datetime, end: datetime):
        """Based on the start time and end time of a file, returns the applicable time (date)

        Parameters
        ----------
        start : datetime
            Start of the applicable time range
        end : datetime
            End of the applicable time range

        Returns
        -------
        : date
            The date of the mean time between start and end
        """
        # In all production processing cases, utc_start and utc_end should be midnight on consecutive days
        # The applicable date is considered to be the mean between the two, ignoring hours, minutes, seconds
        t_0 = start
        t_1 = end
        t_mean = t_0 + 0.5 * (t_1 - t_0)
        applicable_date = datetime.date(t_mean)
        return applicable_date

    def regex_match(self, path: str or Path or CloudPath):
        """Parse and validate a given path against class-attribute defined regex

        Returns
        -------
        : dict
            Match group dict of filename parts
        """
        # AnyPath is polymorphic but self.path will always be a CloudPath or Path object with a name attribute.
        match = self._regex.match(path.name)  # pylint: disable=no-member
        if not match:
            raise ValueError(f"Proposed path {path} failed validation against regex pattern {self._regex}")
        return match.groupdict()

    def generate_prefixed_path(self, parent_path: Union[str, Path, S3Path]) -> Union[Path, S3Path]:
        """Generates an absolute path of the form {parent_path}/{prefix_structure}/{file_basename}
        The parent_path can be an S3 bucket or an absolute local filepath (must start with /)

        Parameters
        ----------
        parent_path : str or Path or S3Path
            Absolute path to the parent directory or S3 bucket prefix. The generated path prefix is appended to the
            parent path and followed by the file basename.

        Returns
        -------
        : Path or S3Path
        """
        if isinstance(parent_path, str):
            parent_path = AnyPath(parent_path)

        if not parent_path.is_absolute():
            raise ValueError(f"Detected relative parent_path {parent_path} passed to generate_prefixed_path. "
                             "The parent_path must be an absolute path. e.g. s3://my-bucket or /starts/with/root.")

        return parent_path / self.archive_prefix / self.path.name


class L0Filename(AbstractValidFilename):
    """Filename validation class for L0 files from EDOS."""

    _regex = LIBERA_L0_REGEX
    _fmt = "{id_char}{scid:03}{first_apid:04}{fill:A<14}{created_time}{numeric_id}{file_number:02}.{extension}{signal}"
    _required_parts = (
        'id_char', 'scid', 'first_apid', 'fill', 'created_time', 'numeric_id', 'file_number', 'extension')

    @property
    def archive_prefix(self):
        """Property that contains the generated prefix for L0 archiving"""
        # Generate prefix structure
        l0_file_type = "CR" if self.filename_parts.file_number == 0 else "PDS"  # CR is always PDS file_number 0
        apid = self.filename_parts.first_apid

        # 2023-07-14: This prefix might become too large over the course of the Libera mission
        return f"{l0_file_type}/{apid:0>4}"

    @classmethod
    def from_filename_parts(cls,  # pylint: disable=arguments-differ
                            basepath: str or Path = None,
                            id_char: str = None,
                            scid: int = None,
                            first_apid: int = None,
                            fill: str = None,
                            created_time: datetime = None,
                            numeric_id: int = None,
                            file_number: int = None,
                            extension: str = None,
                            signal: str = None):
        """Create instance from filename parts. All keyword arguments other than basepath are required!

        The part names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath or prefix.
        id_char : str
            Either P (for PDS files, Construction Records) or X (for Delivery Records)
        scid : int
            Spacecraft ID
        first_apid : int
            First APID in the file
        fill : str
            Custom string up to 14 characters long
        created_time : datetime
            Creation time of the file
        numeric_id : int
            Data set ID, 0-9, one digit
        file_number : str
            File number within the data set. Construction records are always file number zero.
        extension : str
            File name extension. Either PDR or PDS
        signal : str or None, Optional
            Optional signal suffix. Always '.XFR'

        Returns
        -------
        : cls
        """
        cls._check_required_parts(locals())
        return cls._from_filename_parts(basepath=basepath,
                                        id_char=id_char,
                                        scid=scid,
                                        first_apid=first_apid,
                                        fill=fill,
                                        created_time=created_time,
                                        numeric_id=numeric_id,
                                        file_number=file_number,
                                        extension=extension,
                                        signal=signal)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               id_char: str,
                               scid: int,
                               first_apid: int,
                               fill: str,
                               created_time: datetime,
                               numeric_id: int,
                               file_number: int,
                               extension: str,
                               signal: str = None):
        """Construct a path from filename parts

        Parameters
        ----------
        id_char : str
            Either P (for PDS files, Construction Records) or X (for Delivery Records)
        scid : int
            Spacecraft ID
        first_apid : int
            First APID in the file
        fill : str
            Custom string up to 14 characters long
        created_time : datetime
            Creation time of the file
        numeric_id : int
            Data set ID, 0-9, one digit
        file_number : str
            File number within the data set. Construction records are always file number zero.
        extension : str
            File name extension. Either PDR or PDS
        signal : str or None, Optional
            Optional signal suffix. Always '.XFR'

        Returns
        -------
        : str
            Formatted filename
        """
        signal = signal if signal else ""

        return cls._fmt.format(id_char=id_char,
                               scid=scid,
                               first_apid=first_apid,
                               fill=fill,
                               created_time=created_time.strftime(NUMERIC_DOY_TS_FORMAT),
                               numeric_id=numeric_id,
                               file_number=file_number,
                               extension=extension,
                               signal=signal)

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['scid'] = int(d['scid'])
        d['first_apid'] = int(d['first_apid'])
        d['numeric_id'] = int(d['numeric_id'])
        d['file_number'] = int(d['file_number'])
        d['created_time'] = datetime.strptime(d['created_time'], NUMERIC_DOY_TS_FORMAT)
        return SimpleNamespace(**d)


class LiberaDataProductFilename(AbstractValidFilename):
    """Filename validation class for L1B and L2 science products"""

    _regex = LIBERA_DATA_PRODUCT_REGEX
    _fmt = "LIBERA_{data_level}_{product_name}_{version}_{utc_start}_{utc_end}_{revision}.{extension}"
    _required_parts = ('data_level', 'product_name', 'version', 'utc_start', 'utc_end', 'revision', 'extension')

    @property
    def archive_prefix(self):
        """Property that contains the generated prefix for L1B and L2 archiving"""
        # Generate prefix structure
        # <product_type>/<year>/<month>/<day>
        product_name = self.filename_parts.product_name

        applicable_date = self._calculate_applicable_time(self.filename_parts.utc_start, self.filename_parts.utc_end)

        return f"{product_name}/{applicable_date.year:0>4}/{applicable_date.month:0>2}/{applicable_date.day:0>2}"

    @classmethod
    def from_filename_parts(cls,  # pylint: disable=arguments-differ
                            basepath: str or Path = None,
                            data_level: str = None,
                            product_name: str = None,
                            version: str = None,
                            utc_start: datetime = None,
                            utc_end: datetime = None,
                            revision: datetime = None,
                            extension: str = 'nc'):
        """Create instance from filename parts. All keyword arguments other than basepath are required!

        The part names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath or prefix.
        data_level : str
            L1B or L2 identifying the level of the data product
        product_name : str
            Product type. e.g. cloud-fraction for L2 or cam for L1B. May contain anything except for underscores.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        utc_start : datetime
            First timestamp in the SPK
        utc_end : datetime
            Last timestamp in the SPK
        revision: datetime
            Time when the file was created.
        extension : str
            File extension (.nc or .h5)

        Returns
        -------
        : cls
        """
        cls._check_required_parts(locals())
        return cls._from_filename_parts(basepath=basepath,
                                        data_level=data_level,
                                        product_name=product_name,
                                        version=version,
                                        utc_start=utc_start,
                                        utc_end=utc_end,
                                        revision=revision,
                                        extension=extension)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               data_level: str,
                               product_name: str,
                               version: str,
                               utc_start: datetime,
                               utc_end: datetime,
                               revision: datetime,
                               extension: str):
        """Construct a path from filename parts

        Parameters
        ----------
        data_level : str
            L1B or L2
        product_name : str
            Libera instrument, cam or rad for L1B and cloud-fraction etc. for L2. May contain anything except
            for underscores.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        utc_start : datetime
            First timestamp in the SPK
        utc_end : datetime
            Last timestamp in the SPK
        revision: datetime
            Time when the file was created.
        extension : str
            File extension (.nc or .h5)

        Returns
        -------
        : str
            Formatted filename
        """
        return cls._fmt.format(data_level=data_level.upper(),
                               product_name=product_name.upper(),
                               version=version.upper(),
                               utc_start=utc_start.strftime(PRINTABLE_TS_FORMAT),
                               utc_end=utc_end.strftime(PRINTABLE_TS_FORMAT),
                               revision=revision.strftime(REVISION_TS_FORMAT),
                               extension=extension)

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['utc_start'] = datetime.strptime(d['utc_start'], PRINTABLE_TS_FORMAT)
        d['utc_end'] = datetime.strptime(d['utc_end'], PRINTABLE_TS_FORMAT)
        d['revision'] = datetime.strptime(d['revision'], REVISION_TS_FORMAT)
        return SimpleNamespace(**d)


class ManifestFilename(AbstractValidFilename):
    """Class for naming manifest files"""

    _regex = MANIFEST_FILE_REGEX
    _fmt = "LIBERA_{manifest_type}_MANIFEST_{created_time}.json"
    _required_parts = ('manifest_type', 'created_time')

    @property
    def archive_prefix(self):
        """Manifests are not archived like data products, but for convenience and ease of debugging they will be kept
        in the dropbox bucket by input/output and day they were made. This is used by the step function clean up
        function in the CDK.
        # Generate prefix structure
        # <manifest_type>/<year>/<month>/<day>
        """
        manifest_type = self.filename_parts.manifest_type.value

        # This is taking the average of the same time which is the same time
        applicable_date = self.filename_parts.created_time

        return f"{manifest_type}/{applicable_date.year:0>4}/{applicable_date.month:0>2}/{applicable_date.day:0>2}"

    @classmethod
    def from_filename_parts(cls,  # pylint: disable=arguments-differ
                            basepath: str or Path = None,
                            manifest_type: ManifestType = None,
                            created_time: datetime = None):
        """Create instance from filename parts. All keyword arguments other than basepath are required!

        The part names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath or prefix.
        manifest_type : ManifestType
            Input or output
        created_time : datetime
            Time of manifest creation (writing).

        Returns
        -------
        : cls
        """
        cls._check_required_parts(locals())
        return cls._from_filename_parts(basepath=basepath,
                                        manifest_type=manifest_type,
                                        created_time=created_time)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               manifest_type: ManifestType,
                               created_time: datetime):
        """Construct a path from filename parts

        Parameters
        ----------
        manifest_type : ManifestType
            Input or output
        created_time : datetime
            Time of manifest creation (writing).

        Returns
        -------
        : str
            Formatted filename
        """
        return cls._fmt.format(manifest_type=manifest_type.value.upper(),
                               created_time=created_time.strftime(PRINTABLE_TS_FORMAT))

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['manifest_type'] = ManifestType(d['manifest_type'].upper())
        d['created_time'] = datetime.strptime(d['created_time'], PRINTABLE_TS_FORMAT)
        return SimpleNamespace(**d)


class EphemerisKernelFilename(AbstractValidFilename):
    """Class to construct, store, and manipulate an SPK filename"""

    _regex = SPK_REGEX
    _fmt = "LIBERA_{spk_object}_{version}_{utc_start}_{utc_end}_{revision}.bsp"
    _required_parts = ('spk_object', 'version', 'utc_start', 'utc_end', 'revision')

    @property
    def archive_prefix(self):
        """Property that contains the generated prefix for SPICE archiving"""
        # Generate prefix structure
        # <type>/<year>/<month>/<day>
        spk_object = self.filename_parts.spk_object

        applicable_date = self._calculate_applicable_time(self.filename_parts.utc_start, self.filename_parts.utc_end)

        return f"{spk_object}/{applicable_date.year:0>4}/{applicable_date.month:0>2}/{applicable_date.day:0>2}"

    @classmethod
    def from_filename_parts(cls,  # pylint: disable=arguments-differ
                            basepath: str or Path = None,
                            spk_object: str = None,
                            version: str = None,
                            utc_start: datetime = None,
                            utc_end: datetime = None,
                            revision: datetime = None):
        """Create instance from filename parts. All keyword arguments other than basepath are required!

        The part arg names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath or prefix.
        spk_object : str
            Name of object whose attitude is represented in this SPK.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        utc_start : datetime
            Start time of data.
        utc_end : datetime
            End time of data.
        revision: datetime
            When the file was last revised.

        Returns
        -------
        : cls
        """
        cls._check_required_parts(locals())
        return cls._from_filename_parts(basepath=basepath,
                                        spk_object=spk_object,
                                        version=version,
                                        utc_start=utc_start,
                                        utc_end=utc_end,
                                        revision=revision)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               spk_object: str,
                               version: str,
                               utc_start: datetime,
                               utc_end: datetime,
                               revision: datetime):
        """Create an instance from a given path

        Parameters
        ----------
        spk_object : str
            Name of object whose ephemeris is represented in this SPK.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        utc_start : datetime
            Start time of data.
        utc_end : datetime
            End time of data.
        revision: datetime
            Time when the file was last revised

        Returns
        -------
        : cls
        """
        return cls._fmt.format(spk_object=spk_object.upper(),
                               version=version.upper(),
                               utc_start=utc_start.strftime(PRINTABLE_TS_FORMAT),
                               utc_end=utc_end.strftime(PRINTABLE_TS_FORMAT),
                               revision=revision.strftime(REVISION_TS_FORMAT))

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['utc_start'] = datetime.strptime(d['utc_start'], PRINTABLE_TS_FORMAT)
        d['utc_end'] = datetime.strptime(d['utc_end'], PRINTABLE_TS_FORMAT)
        d['revision'] = datetime.strptime(d['revision'], REVISION_TS_FORMAT)
        return SimpleNamespace(**d)


class AttitudeKernelFilename(AbstractValidFilename):
    """Class to construct, store, and manipulate an SPK filename"""

    _regex = CK_REGEX
    _fmt = "LIBERA_{ck_object}_{version}_{utc_start}_{utc_end}_{revision}.bc"
    _required_parts = ('ck_object', 'version', 'utc_start', 'utc_end', 'revision')

    @property
    def archive_prefix(self):
        """Property that contains the generated prefix for SPICE archiving"""
        # Generate prefix structure
        # <type>/<year>/<month>/<day>
        ck_object = self.filename_parts.ck_object

        applicable_date = self._calculate_applicable_time(self.filename_parts.utc_start, self.filename_parts.utc_end)

        return f"{ck_object}/{applicable_date.year:0>4}/{applicable_date.month:0>2}/{applicable_date.day:0>2}"

    @classmethod
    def from_filename_parts(cls,  # pylint: disable=arguments-differ
                            basepath: str or Path = None,
                            ck_object: str = None,
                            version: str = None,
                            utc_start: datetime = None,
                            utc_end: datetime = None,
                            revision: datetime = None):
        """Create instance from filename parts. All keyword arguments other than basepath are required!

        The part arg names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath or prefix.
        ck_object : str
            Name of object whose attitude is represented in this CK.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        utc_start : datetime
            Start time of data.
        utc_end : datetime
            End time of data.
        revision: datetime
            When the file was last revised.

        Returns
        -------
        : cls
        """
        cls._check_required_parts(locals())
        return cls._from_filename_parts(basepath=basepath,
                                        ck_object=ck_object,
                                        version=version,
                                        utc_start=utc_start,
                                        utc_end=utc_end,
                                        revision=revision)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               ck_object: str,
                               version: str,
                               utc_start: datetime,
                               utc_end: datetime,
                               revision: datetime):
        """Create an instance from a given path

        Parameters
        ----------
        ck_object : str
            Name of object whose attitude is represented in this CK.
        utc_start : datetime
            Start time of data.
        utc_end : datetime
            End time of data.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        revision: datetime
            When the file was last revised.

        Returns
        -------
        : cls
        """
        return cls._fmt.format(ck_object=ck_object.upper(),
                               version=version.upper(),
                               utc_start=utc_start.strftime(PRINTABLE_TS_FORMAT),
                               utc_end=utc_end.strftime(PRINTABLE_TS_FORMAT),
                               revision=revision.strftime(REVISION_TS_FORMAT))

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['utc_start'] = datetime.strptime(d['utc_start'], PRINTABLE_TS_FORMAT)
        d['utc_end'] = datetime.strptime(d['utc_end'], PRINTABLE_TS_FORMAT)
        d['revision'] = datetime.strptime(d['revision'], REVISION_TS_FORMAT)
        return SimpleNamespace(**d)


def get_current_revision_str():
    """Get the current `r%y%j%H%M%S` string for filename revisions.

    Returns
    -------
    : str
        Current (now) revision string.
    """
    return datetime.utcnow().strftime(REVISION_TS_FORMAT)


def format_semantic_version(semantic_version: str):
    """Formats a semantic version string X.Y.Z into a filename-compatible string like VX-Y-Z, for X = major version,
    Y = minor version, Z = patch.

    Result is uppercase.
    Release candidate suffixes are allowed as no strict checking is done on the contents of X, Y, or Z.
    e.g. 1.2.3rc1 becomes V1-2-3RC1

    Parameters
    ----------
    semantic_version : str
        String matching X.Y.Z where X, Y and Z are integers of any length

    Returns
    -------
    : str
    """
    major, minor, patch = semantic_version.split('.')
    return f"V{major}-{minor}-{patch}".upper()


def get_current_version_str(package_name: str):
    """Retrieve the current version of a (algorithm) package and format it for inclusion in a filename

    Parameters
    ----------
    package_name : str
        Package for which to retrieve a version string. This should be your algorithm package and it must use a
        semantic versioning scheme, configured in project metadata.

    Returns
    -------
    : str
        Version string in format vM1m2p3
    """
    semver = metadata.version(package_name)
    return format_semantic_version(semver)
